@@ -1,26 +1,27 @@
 # Created by julien piccini
 # email : piccini.julien@gmail.com
-# version : 0.0.7
+# version : 0.0.9
 
 import json as _json
+import os
 import time as _time
 from concurrent import futures as _futures
-from copy import deepcopy as _deepcopy
 from pathlib import Path as _Path
-from typing import Union as _Union
-from typing import IO as _IO
-import os
-from aanalytics2 import config
 
 # Non standard libraries
 import jwt as _jwt
 import pandas as _pd
 import requests as _requests
+from copy import deepcopy as _deepcopy
+from typing import IO as _IO, Optional
+from typing import Union as _Union
+
+from aanalytics2 import config
 
 
 def createConfigFile(verbose: object = False) -> None:
-    """
-    This function will create a 'config_admin.json' file where you can store your access data. 
+    """Creates a `config_admin.json` file with the pre-defined configuration format to store
+    the access data in.
     """
     json_data = {
         'org_id': '<orgID>',
@@ -36,18 +37,42 @@ def createConfigFile(verbose: object = False) -> None:
             f" file created at this location : {os.getcwd()}{os.sep}config_admin.json")
 
 
-def importConfigFile(file: str) -> None:
+def _find_path(path: str) -> Optional[_Path]:
+    """Checks if the file denoted by the specified `path` exists and returns the Path object
+    for the file.
+
+    If the file under the `path` does not exist and the path denotes an absolute path, tries
+    to find the file by converting the absolute path to a relative path.
+
+    If the file does not exist with either the absolute and the relative path, returns `None`.
     """
-    This function will read the 'config_admin.json' to retrieve the information to be used by this module. 
+    if _Path(path).exists():
+        return _Path(path)
+    elif path.startswith('/') and _Path('.' + path).exists():
+        return _Path('.' + path)
+    elif path.startswith('\\') and _Path('.' + path).exists():
+        return _Path('.' + path)
+    else:
+        return None
+
+
+def importConfigFile(path: str) -> None:
+    """Reads the file denoted by the supplied `path` and retrieves the configuration information
+    from it.
+
     Arguments:
-        file: REQUIRED : file (if in the same folder) or path to the file and file if in another folder.
+        path: REQUIRED : path to the configuration file. Can be either a fully-qualified or relative.
+
     Example of file value.
     "config.json"
     "./config.json"
+    "/my-folder/config.json"
     """
-    if file.startswith('/'):
-        file = "."+file
-    with open(_Path(file), 'r') as file:
+    config_file_path: Optional[_Path] = _find_path(path)
+    if config_file_path is None:
+        raise FileNotFoundError(
+            f"Unable to find the configuration file under path `{path}`.")
+    with open(config_file_path, 'r') as file:
         f = _json.load(file)
         config.org_id = f['org_id']
         config.api_key = f['api_key']
@@ -58,18 +83,24 @@ def importConfigFile(file: str) -> None:
 
 
 def retrieveToken(verbose: bool = False, save: bool = False, **kwargs) -> str:
-    """ Retrieve the token by using the information provided by the user during the import importConfigFile function. 
+    """Retrieves the token by using the information provided by the user during
+    the import importConfigFile function.
+
     Argument : 
         verbose : OPTIONAL : Default False. If set to True, print information.
         save : OPTIONAL : Default False. If set to True, will save the token in a txt file (token.txt). 
     """
-    if config.pathToKey.startswith('/'):
-        config.pathToKey = "."+config.pathToKey
-    with open(_Path(config.pathToKey), 'r') as f:
+    private_key_path: Optional[_Path] = _find_path(config.pathToKey)
+    if private_key_path is None:
+        raise FileNotFoundError(
+            f"Unable to find the private key under the path `{config.pathToKey}`.")
+    with open(private_key_path, 'r') as f:
         private_key_unencrypted = f.read()
-        header_jwt = {'cache-control': 'no-cache',
-                      'content-type': 'application/x-www-form-urlencoded'}
-    jwtPayload = {
+    header_jwt = {
+        'cache-control': 'no-cache',
+        'content-type': 'application/x-www-form-urlencoded'
+    }
+    jwt_payload = {
         # Expiration set to 24 hours
         "exp": round(24 * 60 * 60 + int(_time.time())),
         "iss": config.org_id,  # org_id
@@ -78,18 +109,17 @@ def retrieveToken(verbose: bool = False, save: bool = False, **kwargs) -> str:
         "aud": "https://ims-na1.adobelogin.com/c/" + config.api_key
     }
     encoded_jwt = _jwt.encode(
-        jwtPayload, private_key_unencrypted, algorithm='RS256')  # working algorithm
+        jwt_payload, private_key_unencrypted, algorithm='RS256')
     payload = {
         "client_id": config.api_key,
         "client_secret": config.secret,
         "jwt_token": encoded_jwt.decode("utf-8")
     }
-    TokenEndpoint = "https://ims-na1.adobelogin.com/ims/exchange/jwt"
-    response = _requests.post(TokenEndpoint,
-                              headers=header_jwt, data=payload)
+    token_endpoint = "https://ims-na1.adobelogin.com/ims/exchange/jwt"
+    response = _requests.post(token_endpoint, headers=header_jwt, data=payload)
     json_response = response.json()
     token = json_response['access_token']
-    config.header["Authorization"] = "Bearer "+token
+    config.header["Authorization"] = "Bearer " + token
     expire = json_response['expires_in']
     config.date_limit = _time.time() + expire / 1000 - \
         500  # end of time for the token
@@ -97,20 +127,21 @@ def retrieveToken(verbose: bool = False, save: bool = False, **kwargs) -> str:
         with open('token.txt', 'w') as f:  # save the token
             f.write(token)
         if verbose:
-            print('token valid till : ' +
-                  _time.ctime(_time.time() + expire / 1000))
+            print(
+                f"token valid till : {_time.ctime(_time.time() + expire / 1000)}")
             print(f"token has been saved here: {os.getcwd()}{os.sep}token.txt")
     return token
 
 
 def _checkToken(func):
     """decorator that checks that the token is valid before calling the API"""
+
     def checking(*args, **kwargs):  # if function is not wrapped, will fire
         now = _time.time()
         if now > config.date_limit - 1000:
             config.token = retrieveToken(*args, **kwargs)
             if kwargs.get("headers", None) is not None:
-                kwargs['headers']['Authorization'] = "Bearer "+config.token
+                kwargs['headers']['Authorization'] = "Bearer " + config.token
             return func(*args, **kwargs)
         else:  # need to return the function for decorator to return something
             return func(*args, **kwargs)
@@ -236,7 +267,6 @@ def getCompanyId(infos: str = 'all'):
 
 
 class Analytics:
-
     # Endpoints
     header = {"Accept": "application/json",
               "Content-Type": "application/json",
@@ -244,7 +274,6 @@ class Analytics:
               "X-Api-Key": ""
               }
     _endpoint = 'https://analytics.adobe.io/api'
-    _endpoint_report = '/companies/'
     _getRS = '/collections/suites'
     _getDimensions = '/dimensions'
     _getMetrics = '/metrics'
@@ -266,7 +295,7 @@ class Analytics:
         if token is None:
             raise AttributeError(
                 'Expected "token" to be referenced.\nPlease ensure you pass the token.')
-        self.header['Authorization'] = "Bearer "+token
+        self.header['Authorization'] = "Bearer " + token
 
     def getReportSuites(self, txt: str = None, rsid_list: str = None, limit: int = 100, extended_info: bool = False,
                         save: bool = False) -> list:
@@ -328,9 +357,160 @@ class Analytics:
             df_rsids.to_csv('RSIDS.csv', sep='\t')
         if nb_error > 0 or nb_empty > 0:
             print(
-                f'WARNING : Retrieved data are partial.\n{nb_error}/{len(list_urls)+1} requests returned an error.\n{nb_empty}/{len(list_urls)} requests returned an empty response. \nTry to use filter to retrieve reportSuite or increase limit per request')
+                f'WARNING : Retrieved data are partial.\n{nb_error}/{len(list_urls) + 1} requests returned an error.\n{nb_empty}/{len(list_urls)} requests returned an empty response. \nTry to use filter to retrieve reportSuite or increase limit per request')
         return df_rsids
 
+    def getVirtualReportSuites(self, extended_info: bool = False, limit: int = 100, filterIds: str = None, idContains: str = None, segmentIds: str = None, save: bool = True)->list:
+        """
+        return a lit of virtual reportSuites and their id. It can contain more information if expansion is selected.
+        Arguments:
+            extended_info : OPTIONAL : boolean to retrieve the maximum of information.
+            limit : OPTIONAL : How many reportSuite retrieves per serverCall
+            filterIds : OPTIONAL : comma delimited list of virtual reportSuite ID  to be retrieved.
+            idContains : OPTIONAL : element that should be contained in the Virtual ReportSuite Id
+            segmentIds : OPTIONAL : comma delimited list of segmentId contained in the VRSID
+            save : OPTIONAL : if set to True, it will save the list in a file. (Default False)
+        """
+        expansion_values = "globalCompanyKey,parentRsid,parentRsidName,timezone,timezoneZoneinfo,currentTimezoneOffset,segmentList,description,modified,isDeleted,dataCurrentAsOf,compatibility,dataSchema,sessionDefinition,curatedComponents,type"
+        params = {"limit": limit}
+        if extended_info:
+            params['expansion'] = expansion_values
+        if filterIds is not None:
+            params['filterByIds'] = filterIds
+        if idContains is not None:
+            params['idContains'] = idContains
+        if segmentIds is not None:
+            params['segmentIds'] = segmentIds
+        path = f"{self._endpoint_company}/reportsuites/virtualreportsuites"
+        vrsid = getData(path, params=params, headers=self.header)
+        content = vrsid['content']
+        if not extended_info:
+            list_content = [{'name': item['name'], 'vrsid': item['id']}
+                            for item in content]
+            df_vrsids = _pd.DataFrame(list_content)
+        else:
+            df_vrsids = _pd.DataFrame(content)
+        total_page = vrsid['totalPages']
+        last_page = vrsid['lastPage']
+        if not last_page:  # if last_page =False
+            callsToMake = total_page
+            list_params = [{**params, 'page': page}
+                           for page in range(1, callsToMake)]
+            list_urls = [path for x in range(1, callsToMake)]
+            listheaders = [self.header for x in range(1, callsToMake)]
+            workers = min(10, total_page)
+            with _futures.ThreadPoolExecutor(workers) as executor:
+                res = executor.map(lambda x, y, z: getData(
+                    x, y, headers=z), list_urls, list_params, listheaders)
+            res = list(res)
+            list_data = [val for sublist in [r['content']
+                                             for r in res if 'content' in r.keys()] for val in sublist]
+            nb_error = sum(1 for elem in res if 'error_code' in elem.keys())
+            nb_empty = sum(1 for elem in res if 'content' in elem.keys() and len(
+                elem['content']) == 0)
+            if not extended_info:
+                list_append = [{'name': item['name'], 'vrsid': item['id']}
+                               for item in list_data]
+                df_append = _pd.DataFrame(list_append)
+            else:
+                df_append = _pd.DataFrame(list_data)
+            df_vrsids = df_vrsids.append(df_append, ignore_index=True)
+        if save:
+            df_vrsids.to_csv('VRSIDS.csv', sep='\t')
+        if nb_error > 0 or nb_empty > 0:
+            print(
+                f'WARNING : Retrieved data are partial.\n{nb_error}/{len(list_urls)+1} requests returned an error.\n{nb_empty}/{len(list_urls)} requests returned an empty response. \nTry to use filter to retrieve reportSuite or increase limit per request')
+        return df_vrsids
+
+    def getVirtualReportSuite(self, vrsid: str = None, extended_info: bool = False, format: str = 'df')->object:
+        """
+        return a single virtual report suite ID information as dataframe.
+        Arguments:
+            vrsid : REQUIRED : The virtual reportSuite to be retrieved
+            extended_info : OPTIONAL : boolean to add more information
+            format : OPTIONAL : format of the output. 2 values "df" for dataframe and "raw" for raw json.
+        """
+        if vrsid is None:
+            raise Exception("require a Virtual ReportSuite ID")
+        expansion_values = "globalCompanyKey,parentRsid,parentRsidName,timezone,timezoneZoneinfo,currentTimezoneOffset,segmentList,description,modified,isDeleted,dataCurrentAsOf,compatibility,dataSchema,sessionDefinition,curatedComponents,type"
+        params = {}
+        if extended_info:
+            params['expansion'] = expansion_values
+        path = f"{self._endpoint_company}/reportsuites/virtualreportsuites/{vrsid}"
+        data = getData(path, params=params, headers=self.header)
+        if format == "df":
+            data = _pd.DataFrame({vrsid: data})
+        return data
+
+    def createVirtualReportSuite(self, name: str = None, parentRsid: str = None, segmentList: list = None, dataSchema: str = "Cache", data_dict: dict = None, **kwargs)->dict:
+        """
+        Create a new virtual report suite based on the information provided.
+        Arguments:
+            name : REQUIRED : name of the virtual reportSuite
+            parentRsid : REQUIRED : Parent reportSuite ID for the VRS
+            segmentLists : REQUIRED : list of segment id to be applied on the ReportSuite.
+            dataSchema : REQUIRED : Type of schema used for the VRSID. (default "Cache")
+            data_dict : OPTIONAL : you can pass directly the dictionary.
+        """
+        path = f"{self._endpoint_company}/reportsuites/virtualreportsuites"
+        expansion_values = "globalCompanyKey,parentRsid,parentRsidName,timezone,timezoneZoneinfo,currentTimezoneOffset,segmentList,description,modified,isDeleted,dataCurrentAsOf,compatibility,dataSchema,sessionDefinition,curatedComponents,type"
+        params = {'expansion': expansion_values}
+        if data_dict is None:
+            body = {
+                "name": name,
+                "parentRsid": parentRsid,
+                "segmentList": segmentList,
+                "dataSchema": dataSchema,
+                "description": kwargs.get('description', '')
+            }
+        else:
+            if 'name' not in data_dict.keys() or 'parentRsid' not in data_dict.keys() or 'segmentList' not in data_dict.keys() or 'dataSchema' not in data_dict.keys():
+                raise Exception(
+                    "Missing one or more fundamental keys : name, parentRsid, segmentList, dataSchema")
+            body = data_dict
+        res = postData(path, params=params, data=body, headers=self.header)
+        return res
+
+    def deleteVirtualReportSuite(self, vrsid: str = None)->str:
+        """
+        Delete a Virtual Report Suite based on the id passed.
+        Arguments:
+            vrsid : REQUIRED : The id of the virtual reportSuite to delete.
+        """
+        if vrsid is None:
+            raise Exception("require a Virtual ReportSuite ID")
+        path = f"{self._endpoint_company}/reportsuites/virtualreportsuites/{vrsid}"
+        res = deleteData(path, headers=self.header)
+        return res
+
+    def validateVirtualReportSuite(self, name: str = None, parentRsid: str = None, segmentList: list = None,  dataSchema: str = "Cache", data_dict: dict = None, **kwargs)->dict:
+        """
+        Validate the object to create a new virtual report suite based on the information provided.
+        Arguments:
+            name : REQUIRED : name of the virtual reportSuite
+            parentRsid : REQUIRED : Parent reportSuite ID for the VRS
+            segmentLists : REQUIRED : list of segment ids to be applied on the ReportSuite.
+            dataSchema : REQUIRED : Type of schema used for the VRSID (default : Cache).
+            data_dict : OPTIONAL : you can pass directly the dictionary.
+        """
+        path = f"{self._endpoint_company}/reportsuites/virtualreportsuites/validate"
+        expansion_values = "globalCompanyKey, parentRsid, parentRsidName, timezone, timezoneZoneinfo, currentTimezoneOffset, segmentList, description, modified, isDeleted, dataCurrentAsOf, compatibility, dataSchema, sessionDefinition, curatedComponents, type"
+        if data_dict is None:
+            body = {
+                "name": name,
+                "parentRsid": parentRsid,
+                "segmentList": segmentList,
+                "dataSchema": dataSchema,
+                "description": kwargs.get('description', '')
+            }
+        else:
+            if 'name' not in data_dict.keys() or 'parentRsid' not in data_dict.keys() or 'segmentList' not in data_dict.keys() or 'dataSchema' not in data_dict.keys():
+                raise Exception(
+                    "Missing one or more fundamental keys : name, parentRsid, segmentList, dataSchema")
+            body = data_dict
+        res = postData(path, data=body, headers=self.header)
+        return res
+
     def getDimensions(self, rsid: str, tags: bool = False, save=False, **kwargs) -> object:
         """
         Retrieve the list of dimensions from a specific reportSuite.Shrink columns to simplify output.
@@ -444,11 +624,12 @@ class Analytics:
             df_users.to_csv('users.csv', sep='\t')
         if nb_error > 0 or nb_empty > 0:
             print(
-                f'WARNING : Retrieved data are partial.\n{nb_error}/{len(list_urls)+1} requests returned an error.\n{nb_empty}/{len(list_urls)} requests returned an empty response. \nTry to use filter to retrieve users or increase limit')
+                f'WARNING : Retrieved data are partial.\n{nb_error}/{len(list_urls) + 1} requests returned an error.\n{nb_empty}/{len(list_urls)} requests returned an empty response. \nTry to use filter to retrieve users or increase limit')
         return df_users
 
     def getSegments(self, name: str = None, tagNames: str = None, inclType: str = 'all', rsids_list: list = None,
-                    sidFilter: list = None, extended_info: bool = False, format: str = "df", save: bool = False, verbose: bool = False, **kwargs) -> object:
+                    sidFilter: list = None, extended_info: bool = False, format: str = "df", save: bool = False,
+                    verbose: bool = False, **kwargs) -> object:
         """
         Retrieve the list of segments. Returns a data frame. 
         Arguments:
@@ -511,7 +692,7 @@ class Analytics:
             lastPage = segs['lastPage']
             page_nb += 1
             if verbose and page_nb % 10 == 0:
-                print(f"request #{page_nb/10}")
+                print(f"request #{page_nb / 10}")
         if format == "df":
             segments = _pd.DataFrame(data)
         else:
@@ -556,7 +737,8 @@ class Analytics:
         """
         Method that creates a new segment based on the dictionary passed to it.
         Arguments:
-            segmentJSON : REQUIRED : the dictionary that represents the JSON statement for the segment. 
+            segmentJSON : REQUIRED : the dictionary that represents the JSON statement for the segment.
+            More information at this address <https://adobedocs.github.io/analytics-2.0-apis/#/segments/segments_createSegment>
         """
         if segmentJSON is None:
             print('No segment data has been pushed')
@@ -594,7 +776,8 @@ class Analytics:
                          self._getSegments + '/' + segmentID, headers=self.header)
         return seg
 
-    def getCalculatedMetrics(self, name: str = None, tagNames: str = None, inclType: str = 'all', rsids_list: list = None,
+    def getCalculatedMetrics(self, name: str = None, tagNames: str = None, inclType: str = 'all',
+                             rsids_list: list = None,
                              extended_info: bool = False, save=False, **kwargs) -> object:
         """
         Retrieve the list of calculated metrics. Returns a data frame. 
@@ -651,8 +834,8 @@ class Analytics:
         """
         Method that create a specific calculated based on the dictionary passed to it.
         Arguments:
-            metricJSON : REQUIRED : Calculated Metrics information to create.
-            Required :  name, definition, rsid
+            metricJSON : REQUIRED : Calculated Metrics information to create. (Required: name, definition, rsid)
+            More information can be found at this address https://adobedocs.github.io/analytics-2.0-apis/#/calculatedmetrics/calculatedmetrics_createCalculatedMetric
         """
         if metricJSON is None or type(metricJSON) != dict:
             raise Exception(
@@ -754,7 +937,8 @@ class Analytics:
         """
         data_rows = _deepcopy(data_rows)
         dict_data = {}
-        dict_data = {row.get('value', 'missing_value')                     : row['data'] for row in data_rows}
+        dict_data = {row.get('value', 'missing_value')
+                             : row['data'] for row in data_rows}
         if cols is not None:
             n_metrics = len(cols) - 1
         if item_id:  # adding the itemId in the data returned
@@ -883,11 +1067,11 @@ class Analytics:
                         f.write(_json.dumps(report, indent=4))
                 if verbose:
                     print(
-                        f'% of total elements retrieved. TotalElements: {report.get("totalElements","no data")}')
+                        f'% of total elements retrieved. TotalElements: {report.get("totalElements", "no data")}')
                 return obj  # in case loop happening with empty data, returns empty data
             if verbose and total_elements != 0:
                 print(
-                    f'% of total elements retrieved: {round((count_elements/total_elements)*100,2)} %')
+                    f'% of total elements retrieved: {round((count_elements / total_elements) * 100, 2)} %')
             if last_page == False and n_result != float('inf'):
                 if count_elements >= n_result:
                     last_page = True
